from __future__ import annotations

import json
import aiohttp
import asyncio
import weakref

from typing import (
    Any,
    ClassVar,
    Coroutine,
    Optional,
    Union,
    Dict,
    List,
    Type,
    TypeVar,
    TYPE_CHECKING,
)

from urllib.parse import quote as _uriquote

from . import utils
from .utils import MISSING
from .errors import LoginFailure

if TYPE_CHECKING:
    from .types import item

    from .types.item import LangType

    T = TypeVar('T')
    BE = TypeVar('BE', bound=BaseException)
    MU = TypeVar('MU', bound='MaybeUnlock')
    Response = Coroutine[Any, Any, T]


async def json_or_text(response: aiohttp.ClientResponse) -> Union[Dict[str, Any], str]:
    text = await response.text(encoding='utf-8')

    try:

        if response.headers['content-type'] == 'application/json; charset=utf-8':
            return json.loads(text)

    except KeyError:
        pass

    return text


class Route:
    BASE: ClassVar[str] = 'https://tarkov-market.com/api/v1'

    def __init__(self, method: str, path: str, **parameters: Any):
        self.path: str = path
        self.method: str = method

        url = "{}{}".format(self.BASE, path)

        if parameters:
            url = url.format_map({k: _uriquote(v) if isinstance(v, str) else v for k, v in parameters.items()})

        self.url = url

    @property
    def bucket(self) -> str:
        return f'{self.path}'


class MaybeUnlock:

    def __init__(self, lock: asyncio.Lock) -> None:
        self.lock: asyncio.Lock = lock
        self._unlock: bool = True

    def __enter__(self: MU) -> MU:
        return self

    def defer(self) -> None:
        self._unlock = False

    def __exit__(
        self,
        exc_type: Optional[Type[BE]],
        exc: Optional[BE],
        traceback
    ) -> None:

        if self._unlock:
            self.lock.release()


class HTTPClient:

    def __init__(
        self,
        connector: Optional[aiohttp.BaseConnector] = None,
        *,
        token: str,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop() if loop is None else loop
        self.connector = connector
        self._locks: weakref.WeakValueDictionary = weakref.WeakValueDictionary()
        self.token: str = token
        self.__session: aiohttp.ClientSession = MISSING

    async def request(self, route: Route, **kwargs: Any) -> Any:
        bucket = route.bucket
        method = route.method
        url = route.url

        lock = self._locks.get(bucket)

        if lock is None:
            lock = asyncio.Lock()

            if bucket is not None:
                self._locks[bucket] = lock

        headers: Dict[str, str] = {
            "x-api-key": self.token,
        }

        if 'json' in kwargs:
            headers['Content-Type'] = 'application/json'
            kwargs['data'] = utils._to_json(kwargs.pop('json'))

        kwargs['headers'] = headers

        await lock.acquire()

        with MaybeUnlock(lock) as maybe_lock:

            for tries in range(5):

                try:

                    async with self.__session.request(method, url, **kwargs) as response:
                        data = await json_or_text(response)

                        if 'error' in data:
                            reason = data['error']

                            if reason == 'Access denied':
                                raise LoginFailure(f'{self.token} is Invalid API KEY.')

                            if reason in (
                                'You reach your limit of 300 reqs per minute',
                                'You reach your limit of 5 req per minute'
                            ):
                                maybe_lock.defer()
                                self.loop.call_later(60, lock.release)

                            raise

                        return data

                except OSError as e:

                    if tries < 4 and e.errno in (54, 10054):
                        await asyncio.sleep(1 + tries * 2)
                        continue

                    raise

    async def recreate(self):
        self.__session = aiohttp.ClientSession()

    async def close(self):

        if not self.__session.closed:
            await self.__session.close()

    async def __aenter__(self):

        if self.__session is not MISSING and self.__session.closed:
            await self.recreate()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()
        return

    def get_item_by_name(
        self,
        name: str,
        *,
        lang: Optional[LangType] = None
    ) -> Response[List[item.Item]]:

        if lang:
            payload: Dict[str, str] = {
                "q": name,
                "lang": lang
            }
            r = Route('POST', '/item', json=payload)

        else:
            r = Route('GET', '/item?q={item_name}', item_name=name)

        return self.request(r)

    def get_item_by_uid(self, uid) -> Response[List[item.Item]]:
        r = Route('GET', '/item?uid={uid}', uid=uid)
        return self.request(r)

    def get_all_items(self) -> Response[List[item.Item]]:
        r = Route('GET', '/items/all')
        return self.request(r)

    def get_all_bsg_items(self) -> Response[List[item.BSGItem]]:
        r = Route('GET', '/bsg/items/all')
        return self.request(r)

    def save_json(self):
        r = Route('GET', '/bsg/items/all/download')
        return self.request(r)



from __future__ import annotations

import io
import aiohttp
import asyncio

from os import PathLike
from typing import Any, Dict, Optional, List, Callable, Union, TYPE_CHECKING

from .item import Item, BSGItem
from .http import HTTPClient
from .utils import MISSING

if TYPE_CHECKING:
    from .types.item import BSGItem as BSGItemPayload

__all__ = (
    'Client',
)


class Client:

    def __init__(
        self,
        *,
        token: str,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        **options: Any
    ):
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop() if loop is None else loop
        self.token: str = token

        connector: Optional[aiohttp.BaseConnector] = options.pop('connector', None)
        self.http = HTTPClient(connector, token=token, loop=loop)

        self._ready: asyncio.Event = asyncio.Event()
        self._closed: bool = False
        self._clear()

    def _clear(self) -> None:
        self._items: Dict[str, Item] = {}
        self._bsg_items: Dict[str, BSGItem] = {}
        self._ready.clear()

    def _handle_ready(self) -> None:
        self._ready.set()

    def _add_bsg_item(self, payload: BSGItemPayload):
        item_id = payload['_id']
        self._bsg_items[item_id] = BSGItem(payload)

    def is_ready(self) -> bool:
        return self._ready.is_set()

    def get_item(self, item_name: str) -> Item:
        """
        Returns a item with the given name.

        Parameters
        -----------
        item_name: :class:`str`
            The name to search for.

        Returns
        --------
        Optional[:class:`~tarkov_market.Item`]
            The item or `None` if item not found.
        """

        return self._items.get(item_name)

    def find_items(
            self,
            item_name: Optional[str] = None,
            *,
            check: Callable[[Item], bool] = MISSING
    ) -> List[Item]:

        result = []

        if check is MISSING:

            def check(i: Item):

                if item_name.lower() in i.name.lower() or item_name.lower() in i.short_name.lower():
                    return True

                return False

        for item in self.items:

            if not check(item):
                continue

            result.append(item)

        return result

    def get_bsg_item(self, item_id: str) -> BSGItem:
        return self._bsg_items.get(item_id)

    async def fetch_item(self, item_name: str, lang: Optional[str] = None) -> Item:
        """
        Raises
        --------
        :exc:`.NotFound`
            A Item with this Name does not exist.

        Returns
        --------
        :class:`~tarkov_market.Item`
            The Item you requested.
        """

        async with self.http as http:
            data = await http.get_item_by_name(item_name, lang=lang)

        return Item(http=self.http, payload=data[0])

    async def fetch_items(self, item_name: str, lang: Optional[str] = None) -> List[Item]:
        """|coro|
        Gets a :class:`.Item`.

        Returns
        -------
        :class:`.Item`
            The items you requested.
        """

        async with self.http as http:
            data = http.get_item_by_name(item_name, lang=lang)

        return [Item(http=self.http, payload=d) for d in data]

    async def save_items(
            self,
            fp: Union[io.BufferedIOBase, PathLike],
            *,
            seek_begin: bool = True
    ) -> int:

        data = await self.http.save_json()

        if isinstance(fp, io.BufferedIOBase):
            written = fp.write(data)

            if seek_begin is True:
                fp.seek(0)

            return written

        with open(fp, 'wb') as f:
            return f.write(data)

    async def close(self) -> None:

        if self._closed:
            return

        self._closed = True
        await self.http.close()
        self._ready.clear()

    async def wait_until_ready(self) -> None:
        await self._ready.wait()

    async def setup(self) -> None:
        """|coro|

        Setup HTTPClient.
        """

        async def runner():

            async with self.http as session:
                await session.recreate()

                data = await session.get_all_items()

                for payload in data:
                    item = Item(http=self.http, payload=payload)
                    self._items[item.name] = item

                bsg_item = await session.get_all_bsg_items()

                map(self._add_bsg_item, bsg_item.values())

        future = asyncio.ensure_future(runner(), loop=self.loop)
        await future

    @property
    def items(self) -> List[Item]:
        return list(self._items.values())

    async def __aenter__(self):
        return self.http.__aenter__()

    async def __aexit__(self, *args):
        return self.http.__aexit__(*args)



from __future__ import annotations
from typing import Dict, List, Optional, TYPE_CHECKING, Any, Tuple, Union

if TYPE_CHECKING:
    from aiohttp import ClientResponse, ClientWebSocketResponse

    try:
        from requests import Response

        _ResponseType = Union[ClientResponse, Response]
    except ModuleNotFoundError:
        _ResponseType = ClientResponse


class TarkovMarketException(Exception):
    pass


class HTTPException(TarkovMarketException):

    def __init__(self, response: _ResponseType, message: Optional[Union[str, Dict[str, Any]]]):
        self.response: _ResponseType = response
        self.status: int = response.status
        self.text: str

        if isinstance(message, dict):
            error = message.get('error')

            if error:
                self.text = '{}: {}'.format(self.status, error)

            else:
                self.text = f'{self.status}'

        else:
            self.text = message or ''

        super().__init__(self.text)


class LoginFailure(TarkovMarketException):
    pass



import json
from typing import Any


class _MissingSentinel:
    def __eq__(self, other):
        return False

    def __bool__(self):
        return False

    def __repr__(self):
        return '...'


MISSING: Any = _MissingSentinel()


def _to_json(obj: Any) -> str:
    return json.dumps(obj, separators=(',', ':'), ensure_ascii=True)
